(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2020
*)

module type S = sig
  type t
  val compare: t -> t -> int
  val is_max: t -> bool
  val empty: t
  val extend: t -> Lang.Ast.PP.t -> t
  val pp: Format.formatter -> t -> unit
end

(** Module implementing the "max k cyclic call sites" strategy *)
module CyclicCallSites(Config: sig val max_nesting: int end): S = struct
  open Lang.Ast

  type t = Calls of PP.t list
         | CallsMax of PP.t list
  [@@deriving ord]

  let pp fmt = function
    | Calls calls ->
      begin match calls with
        | [] ->
          Format.pp_print_string fmt "ε"
        | _ ->
          Format.pp_print_list
            ~pp_sep:(fun fmt () -> Format.pp_print_string fmt " # ")
            PP.pp
            fmt
            (List.rev calls)
      end
    | CallsMax calls ->
      begin match calls with
        | [] -> Format.pp_print_char fmt '*'
        | _ ->
          Format.fprintf fmt "%a # *"
            (Format.pp_print_list
               ~pp_sep:(fun fmt () -> Format.pp_print_string fmt " # ")
               PP.pp)
            (List.rev calls)
      end

  let empty = Calls []

  let is_max = function
    | CallsMax _ -> true
    | Calls _ -> false

  let count c calls =
    List.fold_left (fun n c' -> if PP.compare c c' = 0 then n+1 else n) 0 calls

  let extend_gen max calls c = match calls with
    | CallsMax _ -> calls
    | Calls l -> if count c l >= max then CallsMax (c :: l) else Calls (c :: l)

  let extend = extend_gen Config.max_nesting
end

(** Module implementing the "last k call sites" strategy, as in k-CFA *)
module LastCallSites(Config: sig val max_length: int end): S = struct
  open Lang.Ast

  type t = int * PP.t list
  [@@deriving ord]

  let pp fmt (len, calls) =
    if len < Config.max_length
    then begin
      match calls with
      | [] ->
        Format.pp_print_string fmt "ε"
      | _ ->
        Format.pp_print_list
          ~pp_sep:(fun fmt () -> Format.pp_print_string fmt " # ")
          PP.pp
          fmt
          (List.rev calls)
    end
    else begin
      match calls with
      | [] ->
        Format.pp_print_char fmt '*'
      | _ ->
        Format.fprintf fmt "* # %a"
          (Format.pp_print_list
             ~pp_sep:(fun fmt () -> Format.pp_print_string fmt " # ")
             PP.pp)
          (List.rev calls)
    end

  let empty = (0, [])

  let is_max (n, _) =
    n >= Config.max_length

  let rec keep n = function
    | [] -> []
    | x :: l ->
      if n <= 0
      then []
      else x :: keep (n-1) l

  let extend_gen max (n, l) c =
    if n < max
    then (n+1, c :: l)
    else (n, keep n (c ::l))

  let extend = extend_gen Config.max_length
end
