(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2020
*)

open Lang.Ast

module AbsDomain
    (Bools: AbstractDomains.Sigs.BOOL_DOMAIN)
    (Ints: AbstractDomains.Sigs.INT_DOMAIN) : sig
  type t = { bools: Bools.t; ints: Ints.t; closures: closures }
  and closures =
    | Clos of (t Env.t) Map.Make(Fun).t
    | ClosTop
  val compare: t -> t -> int
  val bot: t
  val is_bot: t -> bool
  val top: t
  val singleton: [< `Bool of bool | `Int of int | `Closure of Fun.t * t Env.t] -> t
  val ints: Ints.t -> t
  val bools: Bools.t -> t
  val mem: [< `Bool of bool | `Int of int | `Clos of Fun.t] -> t -> bool
  val leq: t -> t -> bool
  val join: t -> t -> t
  val closures_join_map: t -> (Fun.t -> t Env.t -> t) -> t -> t
  val widen: t -> t -> t
  val meet: t -> t -> t
  val pp: Format.formatter -> t -> unit
end = struct
  module M = Map.Make(Fun)

  type t = { bools: Bools.t; ints: Ints.t; closures: closures }
  and closures =
    | Clos of (t Env.t) M.t
    | ClosTop
  [@@deriving ord]

  let clos_bot = Clos M.empty

  let clos_top = ClosTop

  let bot = { bools = Bools.bot; ints = Ints.bot; closures = clos_bot }

  let top = { bools = Bools.top; ints = Ints.top; closures = clos_top }

  let clos_is_bot = function
    | Clos m -> M.is_empty m
    | ClosTop -> false

  let is_bot { bools; ints; closures } =
    Bools.is_bot bools &&
    Ints.is_bot ints &&
    clos_is_bot closures

  let clos_is_top = function
    | ClosTop -> true
    | Clos _ -> false

  let is_top { bools; ints; closures } =
    Bools.is_top bools &&
    Ints.is_top ints &&
    clos_is_top closures

  let bool_singleton b =
    { bools = Bools.singleton b;
      ints = Ints.bot;
      closures = clos_bot;
    }
  let int_singleton n =
    { bools = Bools.bot;
      ints = Ints.singleton n;
      closures = clos_bot;
    }
  let closure_singleton t env =
    { bools = Bools.bot;
      ints = Ints.bot;
      closures = Clos (M.singleton t env);
    }
  let singleton = function
    | `Bool b -> bool_singleton b
    | `Int n -> int_singleton n
    | `Closure (t, env) -> closure_singleton t env

  let ints i =
    { bools = Bools.bot;
      ints = i;
      closures = clos_bot;
    }

  let bools b =
    { bools = b;
      ints = Ints.bot;
      closures = clos_bot;
    }

  let rec join
      { bools = bools1; ints = ints1; closures = clos1 }
      { bools = bools2; ints = ints2; closures = clos2 } =
    { bools = Bools.join bools1 bools2;
      ints = Ints.join ints1 ints2;
      closures = clos_join clos1 clos2;
    }
  and clos_join c1 c2 = match c1, c2 with
    | ClosTop, _ | _, ClosTop -> clos_top
    | Clos m1, Clos m2 ->
      let m =
        M.union (fun _t env1 env2 ->
            Some
              (Env.union (fun _x cs1 cs2 ->
                   Some (join cs1 cs2)
                 )
                  env1 env2
              )
          )
          m1 m2
      in
      Clos m

  let rec meet
      { bools = bools1; ints = ints1; closures = clos1 }
      { bools = bools2; ints = ints2; closures = clos2 } =
    { bools = Bools.meet bools1 bools2;
      ints = Ints.meet ints1 ints2;
      closures = clos_meet clos1 clos2;
    }
  and clos_meet c1 c2 = match c1, c2 with
    | ClosTop, c | c, ClosTop -> c
    | Clos m1, Clos m2 ->
      let m =
        M.merge (fun _t oenv1 oenv2 -> match oenv1, oenv2 with
            | Some env1, Some env2 ->
              Some
                (Env.union (fun _x cs1 cs2 ->
                     Some (meet cs1 cs2)
                   )
                    env1 env2
                )
            | None, Some _ | Some _, None -> None
            | None, None -> assert false
          )
          m1 m2
      in
      Clos m

  let rec leq
      { bools = bools1; ints = ints1; closures = clos1 }
      { bools = bools2; ints = ints2; closures = clos2 }
    =
    Bools.leq bools1 bools2 &&
    Ints.leq ints1 ints2 &&
    clos_leq clos1 clos2
  and clos_leq c1 c2 = match c1, c2 with
    | _, ClosTop -> true
    | ClosTop, _ -> false
    | Clos m1, Clos m2 ->
      M.for_all (fun t env1 ->
          match M.find_opt t m2 with
          | None -> false
          | Some env2 -> Env.equal leq env1 env2
        )
        m1

  let clos_mem t = function
    | ClosTop -> true
    | Clos m -> M.mem t m
  let mem absval { bools; ints; closures } =
    match absval with
    | `Bool b -> Bools.mem b bools
    | `Int n -> Ints.mem n ints
    | `Clos t -> clos_mem t closures

  let closures_join_map fclos_top fclos r = match r.closures with
    | ClosTop -> fclos_top
    | Clos m ->
      M.fold (fun t c acc -> join (fclos t c) acc) m bot

  let rec clos_height = function
    | ClosTop -> 0
    | Clos m ->
      1 +
      M.fold (fun _ env acc ->
          Env.fold (fun _ { bools = _; ints = _; closures = cs } acc ->
              max acc (clos_height cs)
            )
            env
            acc
        )
        m
        0

  let rec clos_truncate h c =
    if h <= 0
    then clos_top
    else match c with
      | ClosTop -> clos_top
      | Clos m ->
        Clos
          (M.map (fun env ->
               Env.map
                 (fun r ->
                    { r with
                      closures = clos_truncate (h-1) r.closures;
                    })
                 env)
              m)

  let rec widen
      { bools = bools1; ints = ints1; closures = cs1 }
      { bools = bools2; ints = ints2; closures = cs2 } =
    { bools = Bools.widen bools1 bools2;
      ints = Ints.widen ints1 ints2;
      closures = clos_widen cs1 cs2;
    }
  and clos_widen c1 c2 = match c1, c2 with
    | ClosTop, _ | _, ClosTop -> clos_top
    | Clos m1, Clos m2 ->
      let h = clos_height c1 in
      let m =
        M.merge (fun _t oenv1 oenv2 -> match oenv1, oenv2 with
            | Some _, None -> oenv1
            | None, Some env2 ->
              Some (Env.map (fun r2 ->
                  { r2 with
                    closures = clos_truncate (h-1) r2.closures;
                  })
                  env2)
            | Some env1, Some env2 ->
              Some
                (Env.union (fun _x cs1 cs2 ->
                     Some (widen cs1 cs2)
                   )
                    env1 env2
                )
            | None, None -> assert false
          )
          m1 m2
      in
      Clos m

  let rec pp fmt ({ bools; ints; closures } as v) =
    if is_bot v
    then Format.fprintf fmt "⊥"
    else if is_top v
    then Format.fprintf fmt "⊤"
    else
      Format.fprintf fmt
        "{ @[@[bools =@ %a@];@ @[ints =@ %a@];@ @[closures =@ %a@]@] }"
        Bools.pp bools
        Ints.pp ints
        pp_closures closures
  and pp_closures fmt = function
    | ClosTop -> Format.pp_print_string fmt "⊤"
    | Clos m ->
      if M.is_empty m
      then Format.pp_print_string fmt "∅"
      else
        Format.fprintf fmt "@[{%a}@]"
          (Format.pp_print_list
             ~pp_sep:(fun fmt () -> Format.fprintf fmt ",@ ")
             (fun fmt (x, mx) ->
                Format.fprintf fmt "@[%a@ %a@]"
                  (Fun.pp_rec true) x
                  (Env.pp pp) mx
             )
          )
          (M.bindings m)
end

module Analyze
    (Log: Logs.S)
    (Options: sig
       val refine_branches: bool
       val branch_is_context: bool
       val let_is_context: bool
       val no_env_restrict: bool
       val kill_unreachable: bool
       val debug: bool
     end)
    (Calls: Calls.S)
    (Bools: AbstractDomains.Sigs.BOOL_DOMAIN)
    (Ints: AbstractDomains.Sigs.INT_DOMAIN) =
struct

  (** The abstract domain *)
  module Abs = AbsDomain(Bools)(Ints)

  (** Memoization module *)
  module MemoFix = struct
    module AbsEnv = struct
      type t = Abs.t Env.t
      [@@deriving ord]

      let pp = Env.pp Abs.pp
    end

    module CallsTerm = struct
      type t = Calls.t * Term.t
      [@@deriving ord]

      let pp fmt (calls, t) =
        Format.fprintf fmt "@[@[[%a]@]@ %a@]"
          Calls.pp calls
          Term.pp t
    end

    module AbsClosure = struct
      type t = (Calls.t * Term.t) * Abs.t Env.t
      [@@deriving ord]

      let pp fmt ((calls, t), env) =
        Format.fprintf fmt "@[@[[%a]@]@ %a@ %a@]@."
          Calls.pp calls
          Term.pp t
          (Env.pp Abs.pp) env

      let widen ((calls1, t1), env1) ((_calls2, _t2), env2) =
        ((calls1, t1),
         Env.union (fun _ a1 a2 -> Some (Abs.widen a1 a2)) env1 env2)
    end

    include
      Memo.MakeFix(Log)(AbsClosure)(Map2.Make(CallsTerm)(AbsEnv))(Abs)
  end

  (** Helper function to define the memoized fixpoint *)
  let cofix f =
    let curry f calls env t = f ((calls, t), env)
    and curry1 f b calls env t = f b ((calls, t), env) in
    let uncurry2 f g b ((calls, t), env) = f (curry1 g) b calls env t in
    curry @@
    MemoFix.cofix ~debug:Options.debug ~transient:false @@
    uncurry2 f

  (** Meet of two environments *)
  let env_meet ~narrow:_ oenv1 oenv2 = match oenv1, oenv2 with
    | None, _ | _, None -> None
    | Some env1, Some env2 ->
      let exception Inconsistent in
      try
        Some
          (Env.union (fun _x v1 v2 ->
               let v12 = Abs.meet v1 v2 in
               if Abs.is_bot v12
               then raise Inconsistent
               else Some v12
             )
              env1 env2
          )
      with Inconsistent -> None

  (** Join of two environments *)
  let env_join oenv1 oenv2 = match oenv1, oenv2 with
    | None, o | o, None -> o
    | Some env1, Some env2 ->
      Some
        (Env.merge (fun _x ov1 ov2 -> match ov1, ov2 with
             | Some v1, Some v2 -> Some (Abs.join v1 v2)
             | None, _ | _, None -> None
           )
            env1 env2
        )

  (** [backward eval t v] is a backward analyzer, that returns
     necessary informations about the variables of [t], when we know
     it forward evaluated to [v] using [eval]. It is useful to refine
     the knowledge after a branch has been taken in the forward
     analyzer. *)
  let rec backward eval t v =
    if Abs.is_bot v
    then None
    else
      match t.Term.term with
      | Term.Var x -> Some (Env.singleton x v)
      | Term.Int n ->
        let v = Abs.(meet v (ints (Ints.singleton n))) in
        if Abs.is_bot v
        then None
        else Some Env.empty
      | Term.UnknownInt ->
        let v = Abs.(meet v @@ ints Ints.(interval min_int max_int)) in
        if Abs.is_bot v
        then None
        else Some Env.empty
      | Term.Bool b ->
        let v = Abs.(meet v (bools (Bools.singleton b))) in
        if Abs.is_bot v
        then None
        else Some Env.empty
      | Term.UnknownBool ->
        let v = Abs.(meet v (bools Bools.top)) in
        if Abs.is_bot v
        then None
        else Some Env.empty
      | Term.Unop (NEG, t) ->
        backward eval t Abs.(ints @@ Ints.Transfer.neg v.ints)
      | Term.Binop(PLUS, t1, t2) ->
        let v1 = eval t1
        and v2 = eval t2 in
        let env1 =
          backward eval t1 Abs.(ints @@ Ints.Transfer.minus v.ints v2.ints)
        and env2 =
          backward eval t2 Abs.(ints @@ Ints.Transfer.minus v.ints v1.ints)
        in
        env_meet ~narrow:false env1 env2
      | Term.Binop(MINUS, t1, t2) ->
        let v1 = eval t1
        and v2 = eval t2 in
        let env1 =
          backward eval t1 Abs.(ints @@ Ints.Transfer.plus v.ints v2.ints)
        and env2 =
          backward eval t2 Abs.(ints @@ Ints.Transfer.(minus v1.ints v.ints))
        in
        env_meet ~narrow:false env1 env2
      | Term.Binop(LE, t1, t2) ->
        let backward_when b =
          let v1 = eval t1
          and v2 = eval t2 in
          let (ints1, ints2) = Ints.Transfer.le v1.ints v2.ints b in
          let env1 = backward eval t1 (Abs.ints ints1)
          and env2 = backward eval t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.Binop(LT, t1, t2) ->
        let backward_when b =
          let v1 = eval t1
          and v2 = eval t2 in
          let (ints1, ints2) = Ints.Transfer.lt v1.ints v2.ints b in
          let env1 = backward eval t1 (Abs.ints ints1)
          and env2 = backward eval t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.Binop(GE, t1, t2) ->
        let backward_when b =
          let v1 = eval t1
          and v2 = eval t2 in
          let (ints1, ints2) = Ints.Transfer.ge v1.ints v2.ints b in
          let env1 = backward eval t1 (Abs.ints ints1)
          and env2 = backward eval t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.Binop(GT, t1, t2) ->
        let backward_when b =
          let v1 = eval t1
          and v2 = eval t2 in
          let (ints1, ints2) = Ints.Transfer.gt v1.ints v2.ints b in
          let env1 = backward eval t1 (Abs.ints ints1)
          and env2 = backward eval t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.Binop(EQ, t1, t2) ->
        let backward_when b =
          let v1 = eval t1
          and v2 = eval t2 in
          let (ints1, ints2) = Ints.Transfer.eq v1.ints v2.ints b in
          let env1 = backward eval t1 (Abs.ints ints1)
          and env2 = backward eval t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.Binop(NEQ, t1, t2) ->
        let backward_when b =
          let v1 = eval t1
          and v2 = eval t2 in
          let (ints1, ints2) = Ints.Transfer.neq v1.ints v2.ints b in
          let env1 = backward eval t1 (Abs.ints ints1)
          and env2 = backward eval t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.BOrElse (t1, t2) ->
        let if_true =
          if Abs.mem (`Bool true) v
          then
            let env1_t = backward eval t1 (Abs.bools @@ Bools.singleton true)
            and env1_f = backward eval t1 (Abs.bools @@ Bools.singleton false)
            and env2_t = backward eval t2 (Abs.bools @@ Bools.singleton true)
            in
            env_join env1_t (env_meet ~narrow:false env1_f env2_t)
          else
            None
        and if_false =
          if Abs.mem (`Bool false) v
          then
            let env1_f = backward eval t1 (Abs.bools @@ Bools.singleton false)
            and env2_f = backward eval t2 (Abs.bools @@ Bools.singleton false)
            in
            env_meet ~narrow:false env1_f env2_f
          else
            None
        in
        env_join if_true if_false
      | Term.BAndAlso (t1, t2) ->
        let if_true =
          if Abs.mem (`Bool true) v
          then
            let env1_t = backward eval t1 (Abs.bools @@ Bools.singleton true)
            and env2_t = backward eval t2 (Abs.bools @@ Bools.singleton true)
            in
            env_meet ~narrow:false env1_t env2_t
          else
            None
        and if_false =
          if Abs.mem (`Bool false) v
          then
            let env1_t = backward eval t1 (Abs.bools @@ Bools.singleton true)
            and env1_f = backward eval t1 (Abs.bools @@ Bools.singleton false)
            in
            let env2_f = backward eval t2 (Abs.bools @@ Bools.singleton false)
            in
            env_join env1_f (env_meet ~narrow:false env1_t env2_f)
          else
            None
        in
        env_join if_true if_false
      | Term.Binop (TIMES, _, _)
      | Term.Lam _
      | Term.FixLam _
      | Term.If _
      | Term.App _
      | Term.Let _
        -> Some Env.empty

  (** The main abstract evaluator *)
  let eval =
    cofix @@
    fun eval _widen calls env t ->
    let eval calls env t =
      let env =
        if Options.no_env_restrict
        then env
        else Vars.restrict_env (Term.fv t) env
      in
      eval false calls env t
    and eval_widen calls env t =
      let env =
        if Options.no_env_restrict
        then env
        else Vars.restrict_env (Term.fv t) env
      in
      eval (Calls.is_max calls) calls env t
    in
    match t.Term.term with
    | Term.Bool b ->
      Abs.singleton (`Bool b)
    | Term.Lam (xloc, x, t0) ->
      let env = Vars.restrict_env (Term.fv t) env in
      Abs.singleton (`Closure (Fun.Lam (t.id, xloc, x, t0), env))
    | Term.FixLam (floc, f, xloc, x, t0) ->
      let env = Vars.restrict_env (Term.fv t) env in
      Abs.singleton (`Closure (Fun.FixLam (t.id, floc, f, xloc, x, t0), env))
    | Term.UnknownInt ->
      Abs.ints Ints.(interval min_int max_int)
    | Term.UnknownBool ->
      Abs.bools Bools.top
    | Term.If (t1, t2, t3) ->
      let ts1 = eval calls env t1 in
      if Options.kill_unreachable && Abs.is_bot ts1
      then ts1
      else
        let calls' =
          if Options.branch_is_context
          then Calls.extend calls t1.id
          else calls
        in
        let ts2 =
          if Abs.mem (`Bool true) ts1
          then
            if Options.refine_branches
            then
              let learned =
                backward (eval calls env)
                  t1 (Abs.singleton (`Bool true))
              in
              match env_meet ~narrow:true (Some env) learned with
              | None -> Abs.bot
              | Some env' -> eval calls' env' t2
            else eval calls' env t2
          else Abs.bot
        and ts3 =
          if Abs.mem (`Bool false) ts1
          then
            if Options.refine_branches
            then
              let learned =
                backward (eval calls env)
                  t1 (Abs.singleton (`Bool false))
              in
              match env_meet ~narrow:true (Some env) learned with
              | None -> Abs.bot
              | Some env' -> eval calls' env' t3
            else eval calls' env t3
          else Abs.bot
        in
        Abs.join ts2 ts3
    | Term.BOrElse (t1, t2) ->
      let ts1 = eval calls env t1 in
      if Abs.mem (`Bool true) ts1
      then
        if Abs.mem (`Bool false) ts1
        then (* TODO: refine environment of [t2] *)
          Abs.join (Abs.singleton (`Bool true)) (eval calls env t2)
        else Abs.singleton (`Bool true)
      else
      if Abs.mem (`Bool false) ts1
      then (* TODO: refine environment of [t2] *)
        eval calls env t2
      else Abs.bot
    | Term.BAndAlso (t1, t2) ->
      let ts1 = eval calls env t1 in
      if Abs.mem (`Bool false) ts1
      then
        if Abs.mem (`Bool true) ts1
        then (* TODO: refine environment of [t2] *)
          Abs.join (Abs.singleton (`Bool false)) (eval calls env t2)
        else Abs.singleton (`Bool false)
      else
      if Abs.mem (`Bool true) ts1
      then (* TODO: refine environment of [t2] *)
        eval calls env t2
      else Abs.bot
    | Term.Var x ->
      Env.find x env
    | Term.App (t1, t2) ->
      let ts1 = eval calls env t1 in
      if Options.kill_unreachable && Abs.is_bot ts1
      then ts1
      else
        let ts2 = eval calls env t2 in
        if Options.kill_unreachable && Abs.is_bot ts2
        then ts2
        else
          let calls' = Calls.extend calls t.Term.id in
          Abs.closures_join_map
            Abs.top
            (fun t0 env0 ->
               match t0 with
               | Lam (_loc, _xloc, x, t') ->
                 eval_widen
                   calls'
                   (Env.add x ts2 env0)
                   t'
               | FixLam (_loc, _floc, f, _xloc, x, t') ->
                 eval_widen
                   calls'
                   (Env.add x ts2
                      (Env.add f
                         (Abs.singleton
                            (`Closure (t0, (Vars.restrict_env (Fun.fv t0) env0))))
                         env0)
                   )
                   t'
            )
            ts1
    | Term.Let (_xloc, x, t1, t2) ->
      let ts1 = eval calls env t1 in
      if Options.kill_unreachable && Abs.is_bot ts1
      then ts1
      else
        let calls' =
          if Options.let_is_context
          then Calls.extend calls t.Term.id
          else calls
        in
        eval calls' (Env.add x ts1 env) t2
    | Term.Int n ->
      Abs.singleton (`Int n)
    | Term.Binop (op, t1, t2) ->
      let ts1 = eval calls env t1 in
      if Options.kill_unreachable && Abs.is_bot ts1
      then ts1
      else let ts2 = eval calls env t2 in
        if Options.kill_unreachable && Abs.is_bot ts2
        then ts2
        else
          begin
            match op with
            | PLUS ->
              Abs.ints @@ Ints.Transfer.plus ts1.Abs.ints ts2.Abs.ints
            | MINUS ->
              Abs.ints @@ Ints.Transfer.minus ts1.Abs.ints ts2.Abs.ints
            | TIMES ->
              Abs.ints @@ Ints.Transfer.mult ts1.Abs.ints ts2.Abs.ints
            | LE ->
              let ((int1_t, int2_t), (int1_f, int2_f)) =
                (Ints.Transfer.le ts1.Abs.ints ts2.Abs.ints true,
                 Ints.Transfer.le ts1.Abs.ints ts2.Abs.ints false)
              in
              let when_true =
                if Ints.(is_bot int1_t || is_bot int2_t)
                then Bools.bot
                else Bools.singleton true
              and when_false =
                if Ints.(is_bot int1_f || is_bot int2_f)
                then Bools.bot
                else Bools.singleton false
              in
              Abs.bools @@ Bools.join when_true when_false
            | LT ->
              let ((int1_t, int2_t), (int1_f, int2_f)) =
                (Ints.Transfer.lt ts1.Abs.ints ts2.Abs.ints true,
                 Ints.Transfer.lt ts1.Abs.ints ts2.Abs.ints false)
              in
              let when_true =
                if Ints.(is_bot int1_t || is_bot int2_t)
                then Bools.bot
                else Bools.singleton true
              and when_false =
                if Ints.(is_bot int1_f || is_bot int2_f)
                then Bools.bot
                else Bools.singleton false
              in
              Abs.bools @@ Bools.join when_true when_false
            | GE ->
              let ((int1_t, int2_t), (int1_f, int2_f)) =
                (Ints.Transfer.ge ts1.Abs.ints ts2.Abs.ints true,
                 Ints.Transfer.ge ts1.Abs.ints ts2.Abs.ints false)
              in
              let when_true =
                if Ints.(is_bot int1_t || is_bot int2_t)
                then Bools.bot
                else Bools.singleton true
              and when_false =
                if Ints.(is_bot int1_f || is_bot int2_f)
                then Bools.bot
                else Bools.singleton false
              in
              Abs.bools @@ Bools.join when_true when_false
            | GT ->
              let ((int1_t, int2_t), (int1_f, int2_f)) =
                (Ints.Transfer.gt ts1.Abs.ints ts2.Abs.ints true,
                 Ints.Transfer.gt ts1.Abs.ints ts2.Abs.ints false)
              in
              let when_true =
                if Ints.(is_bot int1_t || is_bot int2_t)
                then Bools.bot
                else Bools.singleton true
              and when_false =
                if Ints.(is_bot int1_f || is_bot int2_f)
                then Bools.bot
                else Bools.singleton false
              in
              Abs.bools @@ Bools.join when_true when_false
            | EQ ->
              let ((int1_t, int2_t), (int1_f, int2_f)) =
                (Ints.Transfer.eq ts1.Abs.ints ts2.Abs.ints true,
                 Ints.Transfer.eq ts1.Abs.ints ts2.Abs.ints false)
              in
              let when_true =
                if Ints.(is_bot int1_t || is_bot int2_t)
                then Bools.bot
                else Bools.singleton true
              and when_false =
                if Ints.(is_bot int1_f || is_bot int2_f)
                then Bools.bot
                else Bools.singleton false
              in
              Abs.bools @@ Bools.join when_true when_false
            | NEQ ->
              let ((int1_t, int2_t), (int1_f, int2_f)) =
                (Ints.Transfer.neq ts1.Abs.ints ts2.Abs.ints true,
                 Ints.Transfer.neq ts1.Abs.ints ts2.Abs.ints false)
              in
              let when_true =
                if Ints.(is_bot int1_t || is_bot int2_t)
                then Bools.bot
                else Bools.singleton true
              and when_false =
                if Ints.(is_bot int1_f || is_bot int2_f)
                then Bools.bot
                else Bools.singleton false
              in
              Abs.bools @@ Bools.join when_true when_false
          end
    | Term.Unop (NEG, t1) ->
      let ts1 = eval calls env t1 in
      if Options.kill_unreachable && Abs.is_bot ts1
      then ts1
      else Abs.ints @@ Ints.Transfer.neg ts1.Abs.ints

  (** Evaluator in the empty environment, in the empty calling context *)
  let eval t =
    eval Calls.empty Env.empty t

end
