(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2020
*)

open Cmdliner

let analysis_kind_t =
  let open Arg in
  let open CfaLib.Main.AnalysisKind in
  value
  & vflag Nabla [
    Nabla,
    info ["nabla"]
      ~doc:"Analyzer with widening: implements the ∇-CFA analysis.";
    Global,
    info ["global"]
      ~doc:"Global analyzer: naive implementation of a classical CFA.";
    GlobalIndependentAttribute,
    info ["global-independent-attribute"]
      ~doc:"Global analyzer, with independent attribute abstraction of \
            contexts: same as --global but only one environment is \
            associated to a closure. The bindings in the environment \
            contain sets of abstract values.";
  ]

let integer_domain_t =
  let open Arg in
  let open CfaLib.Main in
  value
  & opt (conv IntegerDomain.(parser, printer)) IntegerDomain.default
  & info ["integer-domain"]
    ~docv:"INTEGER_DOMAIN"
    ~doc:("The abstract domain to interpret sets of integers. Possible values: "
          ^ String.concat ", " IntegerDomain.examples ^ ".")

let call_domain_t =
  let open Arg in
  let open CfaLib.Main in
  value
  & opt (conv CallDomain.(parser, printer)) CallDomain.default
  & info ["call-domain"]
    ~docv:"CALL_DOMAIN"
    ~doc:("The abstract domain to interpret calling contexts. Possible values: "
          ^ String.concat ", " CallDomain.examples ^ ".")

let refine_branches_t = let open Arg in
  value
  & flag
  & info ["refine-branches"]
    ~doc:"Refine the branches of if/then/else by performing a local backward analysis."

let branch_is_context_t = let open Arg in
  value
  & flag
  & info ["branch-is-context"]
    ~doc:"Consider branching as creating a new context."

let let_is_context_t = let open Arg in
  value
  & flag
  & info ["let-is-context"]
    ~doc:"Consider let binding as creating a new context."

let no_env_restrict_t = let open Arg in
  value
  & flag
  & info ["no-env-restrict"]
    ~doc:"Do no restrict environments to only the free variables of a subterm during its analysis (makes the analyses slower, because it hinders some context sharing and memoization)."

let kill_unreachable_t = let open Arg in
  value
  & flag
  & info ["kill-unreachable"]
    ~doc:"Kill the analysis of unreachable parts of the code."

let debug_t = let open Arg in
  value
  & flag
  & info ["debug"]
    ~doc:"Output debug information about fixpoint iterations on stderr."

let share_integers_program_points_t = let open Arg in
  value
  & flag
  & info ["share-integers-program-points"]
    ~doc:"Do share program points for identical integers."

let share_variables_program_points_t = let open Arg in
  value
  & flag
  & info ["share-variables-program-points"]
    ~doc:"Do share program points for identical variables from the same scope."

let options_t : CfaLib.Main.Options.t Term.t = let open Term in
  const (fun
          refine_branches
          branch_is_context
          let_is_context
          no_env_restrict
          kill_unreachable
          debug
          integer_domain
          call_domain
          analysis_kind
          share_integers_program_points
          share_variables_program_points ->
      { CfaLib.Main.Options.refine_branches;
        branch_is_context;
        let_is_context;
        no_env_restrict;
        kill_unreachable;
        debug;
        integer_domain;
        call_domain;
        analysis_kind;
        share_integers_program_points;
        share_variables_program_points;
      })
  $ refine_branches_t
  $ branch_is_context_t
  $ let_is_context_t
  $ no_env_restrict_t
  $ kill_unreachable_t
  $ debug_t
  $ integer_domain_t
  $ call_domain_t
  $ analysis_kind_t
  $ share_integers_program_points_t
  $ share_variables_program_points_t

let file_t = let open Arg in
  required
  & pos 0 (some string) None
  & info [] ~docv:"FILE"
    ~doc:"File to analyse"

let cfa_t = let open Term in
  const CfaLib.Main.cfa_print
  $ options_t
  $ file_t

let cfa_info = let open Term in
  info "cfa"
    ~doc:"Control flow analysis for a simple functional language."

let () =
  Term.(exit @@ eval (cfa_t, cfa_info))
