(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2020-2021
*)

(** Variables *)
module Var = struct
  include String
  let pp = Format.pp_print_string
end

(** Environments, i.e. maps from variables to some values *)
module Env = struct
  include Map.Make(Var)
  let pp pp_elt fmt env =
    Format.fprintf fmt "@[[%a]@]"
      (Format.pp_print_list
         ~pp_sep:(fun fmt () -> Format.fprintf fmt ",@ ")
         (fun fmt (x, a) ->
            Format.fprintf fmt "@[%a ->@ %a@]"
              Var.pp x
              pp_elt a))
      (bindings env)
end

(** Sets of variables *)
module Vars: sig
  type t
  val empty: t
  val mem: Var.t -> t -> bool
  val singleton: Var.t -> t
  val add: Var.t -> t -> t
  val union: t -> t -> t
  val remove: Var.t -> t -> t
  val pp: Format.formatter -> t -> unit
  val restrict_env: t -> 'a Env.t -> 'a Env.t
end = struct
  type t = unit Env.t

  let empty = Env.empty

  let mem = Env.mem

  let singleton x = Env.singleton x ()

  let add x vars =
    Env.add x () vars

  let union m1 m2 = Env.union (fun _ () () -> Some ()) m1 m2

  let remove = Env.remove

  let pp fmt vars = let open Format in
    pp_print_list
      ~pp_sep:(fun fmt () -> fprintf fmt ",@ ")
      (fun fmt (x, ()) -> Var.pp fmt x)
      fmt
      (Env.bindings vars)

  let restrict_env fvs m =
    Env.merge (fun _x o ov -> match o, ov with
        | Some (), Some v -> Some v
        | None, Some _ -> None
        | Some (), None -> assert false
        | None, None -> None
      )
      fvs m
end

(** Binary operators *)
type binop = PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ
[@@deriving ord]

let pp_binop fmt = function
  | PLUS -> Format.pp_print_string fmt "+"
  | MINUS -> Format.pp_print_string fmt "-"
  | TIMES -> Format.pp_print_string fmt "*"
  | LE -> Format.pp_print_string fmt "<="
  | LT -> Format.pp_print_string fmt "<"
  | GE -> Format.pp_print_string fmt ">="
  | GT -> Format.pp_print_string fmt ">"
  | EQ -> Format.pp_print_string fmt "="
  | NEQ -> Format.pp_print_string fmt "<>"

(** Unary operators *)
type unop = NEG
[@@deriving ord]

let pp_unop fmt = function
  | NEG -> Format.pp_print_string fmt "-"

(** Outer syntax *)
module Outer = struct
  type fun_arg = Var.t Location.located

  type parens_style =
    | Parentheses
    | BeginEnd

  (** Expressions *)
  type expr = expr_ Location.located
  and expr_ =
    | Var of Var.t
    | Let of fun_arg * expr * expr
    | App of expr * expr
    | Fun of fun_arg * expr
    | FixFun of fun_arg * fun_arg * expr
    | Parens of parens_style * expr
    | UnknownInt
    | UnknownBool
    | Bool of bool
    | BOrElse of expr * expr
    | BAndAlso of expr * expr
    | IfThenElse of expr * expr * expr
    | Int of int
    | Binop of binop * expr * expr
    | Unop of unop * expr

  (** Declarations *)
  type decl =
    | DeclVal of Var.t Location.located * expr
    | DeclFun of Var.t Location.located * fun_arg list Location.located * expr
    | DeclRecFun of Var.t Location.located * fun_arg list Location.located * expr

  (** Programs *)
  type program = decl list

  let pp_fun_arg fmt {Location.contents = x; _} =
    Var.pp fmt x

  let rec pp_expr fmt loc_expr = pp_expr_ fmt loc_expr.Location.contents
  and pp_expr_ fmt = let open Format in
    function
    | Var x -> Var.pp fmt x
    | Let (x, e1, e2) ->
      fprintf fmt "@[let %a =@ %a@ in@ %a@]"
        pp_fun_arg x
        pp_expr e1
        pp_expr e2
    | App (e1, e2) ->
      fprintf fmt "@[%a@ %a@]"
        pp_expr e1
        pp_expr e2
    | Parens (Parentheses, e) ->
      fprintf fmt "@[(%a)@]"
        pp_expr e
    | Fun (fun_arg, e) ->
      fprintf fmt "@[<hv>fun@ %a ->@ %a@]"
        pp_fun_arg fun_arg
        pp_expr e
    | FixFun (f, fun_arg, e) ->
      fprintf fmt "@[<hv>fixfun@ %a %a@ ->@ %a@]"
        pp_fun_arg f
        pp_fun_arg fun_arg
        pp_expr e
    | Parens (BeginEnd, e) ->
      fprintf fmt "@[begin@ %a@ end@]"
        pp_expr e
    | UnknownInt ->
      fprintf fmt "?int"
    | UnknownBool ->
      fprintf fmt "?bool"
    | Bool b ->
      if b
      then fprintf fmt "true"
      else fprintf fmt "false"
    | BOrElse (e1, e2) ->
      fprintf fmt "@[(%a)@ ||@ (%a)@]"
        pp_expr e1
        pp_expr e2
    | BAndAlso (e1, e2) ->
      fprintf fmt "@[(%a)@ &&@ (%a)@]"
        pp_expr e1
        pp_expr e2
    | IfThenElse (e1, e2, e3) ->
      fprintf fmt "@[if@ %a@ then@ %a@ else@ %a@]"
        pp_expr e1
        pp_expr e2
        pp_expr e3
    | Int n ->
      pp_print_int fmt n
    | Binop (op, e1, e2) ->
      fprintf fmt "@[(%a)@ %a@ (%a)@]"
        pp_expr e1
        pp_binop op
        pp_expr e2
    | Unop (op, e) ->
      fprintf fmt "@[%a@ (%a)@]"
        pp_unop op
        pp_expr e

  let pp_decl fmt = let open Format in
    function
    | DeclVal (x, e) ->
      fprintf fmt "@[let@ %a@ =@ %a@]"
        Var.pp x.Location.contents
        pp_expr e
    | DeclFun (x, {Location.contents = args; _}, e) ->
      fprintf fmt "@[let@ %a@ @[(%a)@]@ =@ %a@]"
        Var.pp x.Location.contents
        (pp_print_list ~pp_sep:(fun fmt () -> fprintf fmt "@ ") pp_fun_arg)
        args
        pp_expr e
    | DeclRecFun (x, {Location.contents = args; _}, e) ->
      fprintf fmt "@[let rec@ %a@ @[(%a)@]@ =@ %a@]"
        Var.pp x.Location.contents
        (pp_print_list ~pp_sep:(fun fmt () -> fprintf fmt "@ ") pp_fun_arg)
        args
        pp_expr e

  let pp_program fmt decls = let open Format in
    fprintf fmt "@[<v>%a@]"
      (pp_print_list
         ~pp_sep:(fun fmt () -> fprintf fmt "@.@.")
         pp_decl
      )
      decls

  exception Unbound of Location.t * Var.t

  (** Checks that all variables of a term are bound *)
  let rec check_term vars { Location.contents = t; loc } = match t with
    | Var x ->
      if not @@ Vars.mem x vars
      then raise @@ Unbound (loc, x)
    | UnknownInt | UnknownBool | Bool _ | Int _ ->
      ()
    | Let (x, t1, t2) ->
      check_term vars t1;
      check_term (Vars.add x.contents vars) t2
    | App (t1, t2) | BOrElse (t1, t2) | BAndAlso (t1, t2)
    | Binop (_, t1, t2) ->
      check_term vars t1;
      check_term vars t2
    | Fun (x, t) ->
      check_term (Vars.add x.contents vars) t
    | FixFun (f, x, t) ->
      check_term (Vars.add f.contents @@ Vars.add x.contents vars) t
    | Parens (_, t) | Unop (_, t) ->
      check_term vars t
    | IfThenElse (t1, t2, t3) ->
      check_term vars t1;
      check_term vars t2;
      check_term vars t3

  (** [add_args args vars] adds the variables of the formal arguments
     [args] to the set of variables [vars] *)
  let add_args args vars =
    List.fold_left (fun acc x -> Vars.add x.Location.contents acc)
      vars args.Location.contents

  (** Checks that all variables of a program are bound *)
  let rec check_program vars = function
    | [] -> ()
    | DeclVal (x, e) :: decls ->
      check_term vars e;
      check_program (Vars.add x.contents vars) decls
    | DeclFun (f, args, e) :: decls ->
      check_term (add_args args vars) e;
      check_program (Vars.add f.contents vars) decls
    | DeclRecFun (f, args, e) :: decls ->
      check_term (Vars.add f.contents @@ add_args args vars) e;
      check_program (Vars.add f.contents vars) decls

  (** [check prog] checks that all variables of the program [prog] are bound *)
  let check prog =
    match check_program Vars.empty prog with
    | () -> Ok ()
    | exception Unbound (loc, x) ->
      let msg = Format.asprintf "Unbound variable: %a" Var.pp x in
      Error (loc, msg)

end

module MapMake(X: sig
    include Map.OrderedType
    include Memo.WITH_PP with type t := t
  end)
= struct
  include Map.Make(X)
  let pp ppval fmt m = let open Format in
    pp_print_list
      ~pp_sep:(fun fmt () -> fprintf fmt ";@ ")
      (fun fmt (x, v) ->
         fprintf fmt "@[(@[%a@])@ -> @[%a@]@]"
           X.pp x
           ppval v)
      fmt
      (bindings m)
end

(** Program points *)
module PP = struct
  type t = int
  let compare = [%derive.ord: int]
  module Map = Map.Make(Int)
  let next_pp = ref 0
  let location_map = ref Map.empty
  let int_map = ref Map.empty

  (** Registers a program point for a location *)
  let register loc =
    let pp = !next_pp in
    incr next_pp;
    location_map := Map.add pp loc !location_map;
    pp

  (** Registers a program point for a boolean. We share the same
     identifier for identical booleans. We take negative numbers for
     booleans. It does not conflict with other identifiers, since they
     are non-negative. *)
  let register_bool = function
    | true -> -1
    | false -> -2

  (** Registers a program point for an integer. We share the same
     identifier for identical integers, when [share_integers = true] *)
  let register_int ~share_integers n =
    if share_integers
    then
      begin match Map.find_opt n !int_map with
        | None ->
          let pp = !next_pp in
          incr next_pp;
          int_map := Map.add n pp !int_map;
          pp
        | Some pp -> pp
      end
    else
      let pp = !next_pp in
      incr next_pp;
      int_map := Map.add n pp !int_map;
      pp

  (** Gets the location map, that assigns resolves program points into
     locations, for terms that are not booleans or integers. *)
  let get_location_map () =
    !location_map

  let pp fmt i =
    match Map.find_opt i !location_map with
    | Some loc -> Format.pp_print_string fmt (Location.to_string_compact "" loc)
    | None -> Format.fprintf fmt "?%i" i
end

(** Terms (inner syntax), annotated with unique program points. *)
module Term: sig
  type t = private { id: PP.t; term: t_ }
  and t_ = private
    | Bool of bool
    | UnknownBool
    | If of t * t * t
    | Var of Var.t
    | Lam of PP.t * Var.t * t
    | FixLam of PP.t * Var.t * PP.t * Var.t * t
    | App of t * t
    | Let of PP.t * Var.t * t * t
    | Int of int
    | UnknownInt
    | Binop of binop * t * t
    | Unop of unop * t
    | BOrElse of t * t
    | BAndAlso of t * t
  val compare: t -> t -> int
  val fv: t -> Vars.t
  val pp_rec: bool -> Format.formatter -> t -> unit
  val pp: Format.formatter -> t -> unit
  val translate: share_integers:bool -> share_variables:bool ->
    Outer.program -> t
end = struct
  type t = { id: PP.t; term: t_ }
  and t_ =
    | Bool of bool
    | UnknownBool
    | If of t * t * t
    | Var of Var.t
    | Lam of PP.t * Var.t * t
    | FixLam of PP.t * Var.t * PP.t * Var.t * t
    | App of t * t
    | Let of PP.t * Var.t * t * t
    | Int of int
    | UnknownInt
    | Binop of binop * t * t
    | Unop of unop * t
    | BOrElse of t * t
    | BAndAlso of t * t

  (** Total order: since we give a unique identifier to each subterm,
     it suffices to compare the identifier. This saves a lot a time. *)
  let compare t1 t2 =
    PP.compare t1.id t2.id

  let rec pp_rec parens fmt { term = t; id = _ } =
    let open Format in
    match t with
    | Bool b -> pp_print_bool fmt b
    | If (t1, t2, t3) ->
      if parens
      then
        fprintf fmt "@[(if@ %a@ then@ %a@ else@ %a)@]"
          (pp_rec false) t1
          (pp_rec false) t2
          (pp_rec false) t3
      else
        fprintf fmt "@[if@ %a@ then@ %a@ else@ %a@]"
          (pp_rec false) t1
          (pp_rec false) t2
          (pp_rec false) t3
    | Var x -> Var.pp fmt x
    | Lam (_xloc, x, t) ->
      if parens
      then
        fprintf fmt "@[(fun@ %a@ ->@ %a)@]"
          Var.pp x
          (pp_rec false) t
      else
        fprintf fmt "@[fun@ %a@ ->@ %a@]"
          Var.pp x
          (pp_rec false) t
    | FixLam (_floc, f, _xloc, x, t) ->
      if parens
      then
        fprintf fmt "@[(rec fun@ %a@ %a@ ->@ %a)@]"
          Var.pp f
          Var.pp x
          (pp_rec false) t
      else
        fprintf fmt "@[rec fun@ %a@ %a@ ->@ %a@]"
          Var.pp f
          Var.pp x
          (pp_rec false) t
    | App (t1, t2) ->
      let parens1 = match t1.term with
        | Bool _ | Var _ | App _ | UnknownInt | UnknownBool | Int _ -> false
        | FixLam _ | Lam _ | Let _ | If _ | Binop _ | Unop _ | BOrElse _
        | BAndAlso _ -> true
      in
      let parens2 = match t2.term with
        | Bool _ | Var _ | UnknownInt | UnknownBool | Int _ -> false
        | App _ | FixLam _ | Lam _ | Let _ | If _ | Binop _ | Unop _
        | BOrElse _ | BAndAlso _ -> true
      in
      if parens
      then
        fprintf fmt "@[(%a@ %a)@]"
          (pp_rec parens1) t1
          (pp_rec parens2) t2
      else
        fprintf fmt "@[%a@ %a@]"
          (pp_rec parens1) t1
          (pp_rec parens2) t2
    | Let (_xloc, x, t1, t2) ->
      if parens
      then
        fprintf fmt "@[(let@ %a@ =@ %a@ in@ %a)@]"
          Var.pp x
          (pp_rec false) t1
          (pp_rec false) t2
      else
        fprintf fmt "@[let@ %a@ =@ %a@ in@ %a@]"
          Var.pp x
          (pp_rec false) t1
          (pp_rec false) t2
    | UnknownBool -> Format.pp_print_string fmt "bool?"
    | UnknownInt -> Format.pp_print_string fmt "int?"
    | Int n -> Format.pp_print_int fmt n
    | Binop (PLUS, t1, t2) ->
      let parens1 = match t1.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _
        | Binop (TIMES, _, _) -> false
        | If _ | Lam _ | FixLam _
        | Let _ | Binop ((PLUS | MINUS | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _  -> true
      and parens2 = match t2.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _
        | If _ | Lam _ | FixLam _
        | Let _
        | Binop ((PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ), _, _) ->
          false
        | Unop _ | BOrElse _ | BAndAlso _ -> true
      and parens2b = match t2.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _
        | Binop (TIMES, _, _) -> false
        | If _ | Lam _ | FixLam _
        | Let _
        | Binop ((PLUS | MINUS | LE | LT | GE | GT | EQ | NEQ), _, _) | Unop _
        | BOrElse _ | BAndAlso _  -> true
      in
      if parens
      then
        fprintf fmt "@[(%a@ +@ %a)@]"
          (pp_rec parens1) t1
          (pp_rec parens2) t2
      else
        fprintf fmt "@[%a@ +@ %a@]"
          (pp_rec parens1) t1
          (pp_rec parens2b) t2
    | Binop (MINUS, t1, t2) ->
      let parens1 = match t1.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _
        | Binop(TIMES, _, _) -> false
        | If _ | Lam _ | FixLam _
        | Let _ | Binop ((PLUS | MINUS | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _  -> true
      and parens2 = match t2.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _
        | If _ | Lam _ | FixLam _
        | Let _ | Binop (TIMES, _, _) -> false
        | Binop ((PLUS | MINUS | LE | LT | GE | GT | EQ | NEQ), _, _) | Unop _
        | BOrElse _ | BAndAlso _ -> true
      and parens2b = match t2.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _ -> false
        | If _ | Lam _ | FixLam _
        | Let _
        | Binop ((PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _ -> true
      in
      if parens
      then
        fprintf fmt "@[(%a@ -@ %a)@]"
          (pp_rec parens1) t1
          (pp_rec parens2) t2
      else
        fprintf fmt "@[%a@ -@ %a@]"
          (pp_rec parens1) t1
          (pp_rec parens2b) t2
    | Binop (TIMES, t1, t2) ->
      let parens1 = match t1.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _ -> false
        | If _ | Lam _ | FixLam _
        | Let _
        | Binop ((PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _  -> true
      and parens2 = match t2.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _
        | If _ | Lam _ | FixLam _
        | Let _ | Binop (TIMES, _, _) -> false
        | Binop ((PLUS | MINUS | LE | LT | GE | GT | EQ | NEQ), _, _) | Unop _
        | BOrElse _ | BAndAlso _ -> true
      and parens2b = match t2.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _ -> false
        | If _ | Lam _ | FixLam _
        | Let _
        | Binop ((PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _ -> true
      in
      if parens
      then
        fprintf fmt "@[(%a@ *@ %a)@]"
          (pp_rec parens1) t1
          (pp_rec parens2) t2
      else
        fprintf fmt "@[%a@ *@ %a@]"
          (pp_rec parens1) t1
          (pp_rec parens2b) t2
    | Binop ((LE | LT | GE | GT | EQ | NEQ) as op, t1, t2) ->
      let parens2b = match t2.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _ -> false
        | If _ | Lam _ | FixLam _
        | Let _
        | Binop ((PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _ -> true
      in
      if parens
      then
        fprintf fmt "@[(%a@ %a@ %a)@]"
          (pp_rec false) t1
          pp_binop op
          (pp_rec false) t2
      else
        fprintf fmt "@[%a@ %a@ %a@]"
          (pp_rec false) t1
          pp_binop op
          (pp_rec parens2b) t2
    | Unop (NEG, t1) ->
      let parens1 = match t1.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ -> false
        | App _ | If _ | Lam _ | FixLam _
        | Let _
        | Binop ((PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _ -> true
      in
      if parens
      then
        fprintf fmt "@[(-%a)@]"
          (pp_rec parens1) t1
      else
        fprintf fmt "@[-%a@]"
          (pp_rec parens1) t1
    | BOrElse (t1, t2) ->
      let parens1 = match t1.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _ ->
          false
        | If _ | Lam _ | FixLam _
        | Let _
        | Binop ((PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _  -> true
      and parens2 = match t2.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _
        | If _ | Lam _ | FixLam _
        | Let _ -> false
        | Binop ((PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _ -> true
      and parens2b = match t2.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _ ->
          false
        | If _ | Lam _ | FixLam _
        | Let _
        | Binop ((PLUS | TIMES | MINUS | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _ -> true
      in
      if parens
      then
        fprintf fmt "@[(%a@ ||@ %a)@]"
          (pp_rec parens1) t1
          (pp_rec parens2) t2
      else
        fprintf fmt "@[%a@ ||@ %a@]"
          (pp_rec parens1) t1
          (pp_rec parens2b) t2
    | BAndAlso (t1, t2) ->
      let parens1 = match t1.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _ ->
          false
        | If _ | Lam _ | FixLam _
        | Let _
        | Binop ((PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _  -> true
      and parens2 = match t2.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _
        | If _ | Lam _ | FixLam _
        | Let _ -> false
        | Binop ((PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _ -> true
      and parens2b = match t2.term with
        | Int _ | UnknownInt | UnknownBool | Bool _ | Var _ | App _ -> false
        | If _ | Lam _ | FixLam _
        | Let _
        | Binop ((PLUS | MINUS | TIMES | LE | LT | GE | GT | EQ | NEQ), _, _)
        | Unop _ | BOrElse _ | BAndAlso _ -> true
      in
      if parens
      then
        fprintf fmt "@[(%a@ &&@ %a)@]"
          (pp_rec parens1) t1
          (pp_rec parens2) t2
      else
        fprintf fmt "@[%a@ &&@ %a@]"
          (pp_rec parens1) t1
          (pp_rec parens2b) t2

  let pp = pp_rec false

  (** Free variables (with memoization) *)
  let fv =
    let module MiniTerm = struct
      type nonrec t = t
      let compare = compare
      let pp = pp
    end in
    let module TMemo =
      Memo.MakeMemo(Logs.Silent)(MiniTerm)(MapMake(MiniTerm))(Vars)
    in
    TMemo.fix ~transient:true @@ fun fv { term = t; id = _ } ->
    match t with
    | Bool _ | UnknownInt | UnknownBool | Int _ -> Vars.empty
    | If (t1, t2, t3) -> Vars.union (fv t1) @@ Vars.union (fv t2) (fv t3)
    | Var x -> Vars.singleton x
    | Lam (_xloc, x, t) -> Vars.remove x (fv t)
    | FixLam (_floc, f, _xloc, x, t) -> Vars.remove f (Vars.remove x (fv t))
    | App (t1, t2) | Binop (_, t1, t2) | BOrElse (t1, t2)
    | BAndAlso (t1, t2) -> Vars.union (fv t1) (fv t2)
    | Let (_xloc, x, t1, t2) -> Vars.union (fv t1) (Vars.remove x (fv t2))
    | Unop (_, t) -> fv t

  let with_loc loc t =
    { id = PP.register loc; term = t }

  (** Translates outer expressions into terms that are annotated with
     unique identifers. Identical variables in the same scope share
     the same identifier, when [share_variables = true]. Identical
     integers also share the same identifier when [share_intergers =
     true]. Identical booleans always share the same identifier. *)
  let rec translate_term ~share_integers ~share_variables
      env { Location.loc; contents } =
    match contents with
    | Outer.Var x ->
      if share_variables
      then { id = Env.find x env; term = Var x }
      else with_loc loc (Var x)
    | Outer.UnknownInt -> with_loc loc UnknownInt
    | Outer.UnknownBool -> with_loc loc UnknownBool
    | Outer.Let (x, t1, t2) ->
      with_loc loc @@
      Let (PP.register x.loc, x.contents,
           translate_term ~share_integers ~share_variables env t1,
           translate_term ~share_integers ~share_variables
             (Env.add x.contents (PP.register x.loc) env) t2)
    | Outer.App (t1, t2) ->
      with_loc loc @@
      App (translate_term ~share_integers ~share_variables env t1,
           translate_term ~share_integers ~share_variables env t2)
    | Outer.Fun (x, t) ->
      with_loc loc @@
      Lam (PP.register x.loc, x.contents,
           translate_term ~share_integers ~share_variables
             (Env.add x.contents (PP.register x.loc) env) t)
    | Outer.FixFun (f, x, t) ->
      with_loc loc @@
      FixLam (PP.register f.loc, f.contents, PP.register x.loc, x.contents,
              translate_term ~share_integers ~share_variables
                (Env.add f.contents (PP.register f.loc) @@
                 Env.add x.contents (PP.register x.loc) @@
                 env)
                t)
    | Outer.Parens (_, t) ->
      translate_term ~share_integers ~share_variables env t
    | Outer.Bool b -> { id = PP.register_bool b; term = Bool b }
    | Outer.BOrElse (t1, t2) ->
      with_loc loc @@
      BOrElse (translate_term ~share_integers ~share_variables env t1,
               translate_term ~share_integers ~share_variables env t2)
    | Outer.BAndAlso (t1, t2) ->
      with_loc loc @@
      BAndAlso (translate_term ~share_integers ~share_variables env t1,
                translate_term ~share_integers ~share_variables env t2)
    | Outer.IfThenElse (t1, t2, t3) ->
      with_loc loc @@
      If (translate_term ~share_integers ~share_variables env t1,
          translate_term ~share_integers ~share_variables env t2,
          translate_term ~share_integers ~share_variables env t3)
    | Outer.Int n ->
      { id = PP.register_int ~share_integers n; term =  Int n }
    | Outer.Binop (op, t1, t2) ->
      with_loc loc @@
      Binop(op,
            translate_term ~share_integers ~share_variables env t1,
            translate_term ~share_integers ~share_variables env t2)
    | Outer.Unop (op, t) ->
      with_loc loc @@
      Unop(op, translate_term ~share_integers ~share_variables env t)

  let rec translate_program = function
    | [] -> assert false
    | [Outer.DeclVal (_, t)] ->
      t
    | Outer.DeclVal (x, t) :: decls ->
      let t1 = t in
      let t2 = translate_program decls in
      let open Location in
      locate (from_pos (get_start x.loc, get_end t2.loc)) @@
      Outer.Let (x, t1, t2)
    | [Outer.DeclFun (_, args, t)] ->
      let open Location in
      List.fold_right
        (fun x acc ->
           locate (from_pos (get_start x.loc, get_end acc.loc))
           @@ Outer.Fun (x, acc)
        )
        args.contents
        t
    | Outer.DeclFun (f, args, t) :: decls ->
      let open Location in
      let t1 =
        List.fold_right
          (fun x acc ->
             locate (from_pos (get_start x.loc, get_end acc.loc))
             @@ Outer.Fun (x, acc)
          )
          args.contents
          t
      in
      let t2 = translate_program decls in
      locate (from_pos (get_start f.loc, get_end t2.loc)) @@
      Outer.Let (f, t1, t2)
    | Outer.DeclRecFun (_f, { contents = []; _ }, _t) :: _ ->
      assert false
    | [Outer.DeclRecFun (f, { contents = x::xs; loc = args_loc }, t)] ->
      let open Location in
      locate (from_pos (get_start args_loc, get_end t.loc)) @@
      Outer.FixFun
        (f,
         x,
         List.fold_right
           (fun x acc ->
              locate (from_pos (get_start x.loc, get_end acc.loc)) @@
              Outer.Fun (x, acc)
           )
           xs
           t
        )
    | Outer.DeclRecFun (f, { contents = x::xs; loc = args_loc }, t) :: decls ->
      let open Location in
      let t1 =
        locate (from_pos (get_start args_loc, get_end t.loc)) @@
        Outer.FixFun
          (f,
           x,
           List.fold_right
             (fun x acc ->
                locate (from_pos (get_start x.loc, get_end acc.loc)) @@
                Outer.Fun (x, acc)
             )
             xs
             t
          )
      in
      let t2 = translate_program decls in
      locate (from_pos (get_start f.loc, get_end t2.loc)) @@
      Outer.Let (f, t1, t2)

  (** Translates programs into terms that are annotated with unique
     identifers. Identical variables in the same scope share the same
     identifier, when [share_variables = true]. Identical integers
     also share the same identifier when [share_intergers = true].
     Identical booleans always share the same identifier. *)
  let translate ~share_integers ~share_variables prog =
    translate_term ~share_integers ~share_variables
      Env.empty (translate_program prog)
end
