(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2020
*)

(** Module signature for types equipped with a pretty-printer *)
module type WITH_PP = sig
  type t
  val pp: Format.formatter -> t -> unit
end

(** Module signature for parameterized types equipped with a pretty-printer *)
module type WITH_PP1 = sig
  type 'a t
  val pp: (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a t -> unit
end

(** A minimalistic module signature for finite maps *)
module type MAP = sig
  type 'a t
  type key
  val empty: 'a t
  val find_opt: key -> 'a t -> 'a option
  val update: key -> ('a option -> 'a option) -> 'a t -> 'a t
  val cardinal: 'a t -> int
  include WITH_PP1 with type 'a t := 'a t
end

(** A larger module signature for finite maps *)
module type MAP2 = sig
  include MAP
  val merge: (key -> 'a option -> 'b option -> 'c option) -> 'a t -> 'b t -> 'c t

  (** [fold_similar f k m acc] folds the function [f] over the
     elements that are "similar" to the key [k] in the map [m],
     starting with accumulator [acc] *)
  val fold_similar: (key -> 'a -> 'b -> 'b) -> key -> 'a t -> 'b -> 'b
end

(** Abstract domain *)
module type DOMAIN = sig
  type t

  (** Bottom element *)
  val bot: t

  (** Pre-order *)
  val leq: t -> t -> bool

  (** Over-approximation of the upper-bound *)
  val join: t -> t -> t

  (** Over-approximation of the upper-bound, that ensures convergence:
     for any sequence [(v_n)_n], the sequence [(u_n)_n], defined as
     follows, must be stationary:
      * [u_0]  is arbitrary
      * [u_(n+1) = widen u_n v_n]
  *)
  val widen: t -> t -> t
  include WITH_PP with type t := t
end

(** Functor to create memoizing functions *)
module MakeMemo
    (Log: Logs.S)(X: WITH_PP)(M: MAP with type key = X.t)(V: WITH_PP):
sig
  val memo: (X.t -> V.t) -> (X.t -> V.t)
  val fix: ?transient:bool -> ?debug:bool ->
    ((X.t -> V.t) -> (X.t -> V.t)) -> (X.t -> V.t)
end
=
struct
  (** Inititializes the memoization table *)
  let init () =
    ref M.empty

  (** Empties the memoization table *)
  let reset r =
    r := M.empty

  (** Gets a value from the cache *)
  let get cache t = M.find_opt t !cache

  (** Updates a value in the cache: requires that a value should be
     already present *)
  let update cache t c =
    cache := M.update t
        (function
          | Some _ -> assert false
          | None -> Some c
        )
        !cache

  (** [memo f] memoizes the [(v, f v)] pairs in a table, so that
     subsequent calls to the memoized [f] are faster *)
  let memo f =
    let cache = init () in
    let g x =
      match get cache x with
      | Some y -> y
      | None ->
        let y = f x in
        update cache x y;
        y
    in
    g

  (** Pretty-printer for the cache *)
  let pp_cache fmt m =
    M.pp V.pp fmt m

  (** [fix ~transient ~debug f] memoizes the functional [f] so that
     intermediate results in the computation of the fixpoint of [f]
     are recorded in a table for faster computation. This function is
     meant to exploit memoization when defining a recursive function.
     If [transient = true], then the table persists, and is also used
     to memoize subsequent calls to the fixpoint. This means that if
     [transient = false], then the cache is not shared between
     different calls to the fixpoint (the cache is emptied at the end
     of each call). If [debug = true], then the contents of the cache
     is printed at the end of the fixpoint computation. *)
  let fix ?(transient=true) ?(debug=false) f =
    let cache = init () in
    let rec self x =
      match get cache x with
      | Some y -> y
      | None ->
        let y = f self x in
        update cache x y;
        y
    in
    (fun x ->
       let res = self x in
       if debug
       then
         Log.eprintf
           "Elements in cache:@.@[%a@]@.Total elements in cache: %i@."
           pp_cache !cache
           (M.cardinal !cache);
       if not transient
       then
         reset cache;
       res)
end

(** Statistics printed by the [MakeFix] module *)
module Statistics = struct
  type t = {
    states: int; (** number of explored states (nodes in the call graph) *)
    edges: int; (** number of explored edges (edges in the call graph) *)
    max_fanout: int; (** maximum fanout degree of states (indicates large disjunctions) *)
    iterations: int; (** number of iterations performed to reach a fixpoint *)
  }

  (** Zero-filled statistics *)
  let zero = {
    states = 0;
    edges = 0;
    max_fanout = 0;
    iterations = 0;
  }

  (** Pretty-printer *)
  let pp fmt { states; edges; max_fanout; iterations } =
    Format.fprintf fmt "states: %i; edges: %i; max fanout: %i; iterations: %i"
      states
      edges
      max_fanout
      iterations
end

(** Functor to compute post-fixpoints of recursively defined functions
   with values in a semi-lattice equipped with widening. *)
module MakeFix
    (Log: Logs.S)
    (X: sig
       include WITH_PP
       val compare: t -> t -> int
       val widen: t -> t -> t
     end)
    (M: MAP2 with type key = X.t)
    (D: DOMAIN) :
sig
  val cofix: ?debug:bool -> ?transient:bool ->
    ((bool -> X.t -> D.t) -> (bool -> X.t -> D.t)) ->
    (X.t -> D.t * Statistics.t)
end
=
struct
  (** type used to indicate whether we might have detected a cycle *)
  type cycle_status = NoCycle | MayCycle

  let join_status s1 s2 = match s1, s2 with
    | NoCycle, s | s, NoCycle -> s
    | MayCycle, MayCycle -> MayCycle

  (** Whether a node has been visited already, and whether it may
     belong to a cycle *)
  type metadata = { visited: bool; status: cycle_status }

  type cache = (metadata * D.t) M.t

  (** Initializes the cache *)
  let init () =
    ref (M.empty : cache)

  (** Empties the cache *)
  let reset r =
    r := (M.empty: cache)

  (** Marks a node as not visited in the local cache: it must not be
     present already in the local cache. The global cache is updated
     so that the node is maked as not visited. *)
  let mark_not_visited local global node =
    local := M.update node
        (function
          | Some _ -> assert false
          | None -> begin match M.find_opt node !global with
              | Some ({ visited = _; status }, value) ->
                Some ({ visited = false; status }, value)
              | None -> Some ({ visited = false; status = NoCycle }, D.bot)
            end
        )
        !local

  (** Gets an entry on the cache *)
  let get cache node = M.find_opt node !cache

  (** Updates the value associated to a node with a new value, and
     marks the node as visited. There must be an entry for that node
     in the cache already. *)
  let update_and_mark_visited cache node new_value =
    cache := M.update node (function
        | Some ({ visited = _; status }, _value) ->
          Some ({ visited = true; status }, new_value)
        | None -> assert false)
        !cache

  (* let pp_cache fmt (m:t) =
   *   M.pp (fun fmt (_meta, v) -> D.pp fmt v) fmt m *)

  module Memo = MakeMemo(Logs.Silent)(X)(M)(X)

  (** Edges are so far only recorded for statistics.
      TODO: use them to track dependencies, and avoid useless recomputations.
  *)
  module Edges = struct
    module Nodes = Set.Make(X)
    module M = Map.Make(X)
    type edges = Nodes.t M.t

    (** Computes how many edges were recorded *)
    let cardinal (edges: edges) =
      M.fold (fun _ ns acc -> Nodes.cardinal ns + acc)
        edges
        0

    (** Computes the maximal fanout degree for the recorded edges *)
    let max_fanout (edges: edges) =
      M.fold (fun _ ns acc -> max (Nodes.cardinal ns) acc)
        edges
        0

    (** Initializes the edges *)
    let init () = ref M.empty

    (** Initializes the edges *)
    let record (edges: edges ref) x1 x2 =
      edges := M.update x1
          (function
            | None -> Some (Nodes.singleton x2)
            | Some xs -> Some (Nodes.add x2 xs)
          )
          !edges
  end

  (** Computes a post-fixpoint of some higher-order functional. *)
  let cofix ?(debug=false) ?(transient=true) f =
    let global : cache ref =
      (** the global cache, that persists across iterations *)
      init ()
    and local : cache ref =
      (** the local cache, that is used inside a single iteration only *)
      init ()
    and edges =
      (** the edges that we have so far recorded *)
      Edges.init ()
    in
    let widen_input =
      (** function that maps some input to a widened version of that input *)
      Memo.memo @@ fun x ->
      match
        M.fold_similar (fun z _ acc ->
            match acc with
            | None -> Some z
            | Some z_acc -> Some (X.widen z_acc z)
          )
          x
          !local
          None
      with
      | Some x' -> X.widen x' x
      | None -> x
    in
    (** one exploration of the graph *)
    let rec explore_from widen x =
      (** explores the graph from the node [x], using [widen] as a
         hint on whether to perform input widening *)
      match get local x with
      | Some ({ visited; status = _ }, y) ->
        if not visited
        then
          (** we query for a node, while its visit is not finished
             yet: we have a cycle *)
          local := M.update x (function _ ->
              Some ({ visited; status = MayCycle }, y))
              !local;
        y
      | None ->
        let x' =
          (** the new input we should consider *)
          if widen
          then
            let x' = widen_input x in
            (** record the widening edge *)
            if X.compare x x' <> 0
            then Edges.record edges x x';
            x'
          else x
        in
        let () = mark_not_visited local global x in
        let explore_from_and_record_calls widen z =
          (** record the call edge *)
          if X.compare x z <> 0
          then Edges.record edges x z;
          explore_from widen z
        in
        let y = f explore_from_and_record_calls widen x' in
        update_and_mark_visited local x y;
        y
    in
    (** iterate the graph explorations until we get a post-fixpoint *)
    let nb_iterations = ref 0 in
    let rec stabilize_from x =
      incr nb_iterations;
      reset local;
      let y = explore_from false x (** start exploring with no widening *)
      and changed = ref false (** tells whether the graph was modified *)
      and nb_states = ref 0
      and nb_new_states = ref 0
      and nb_explored_states = ref 0
      and nb_unstable_states = ref 0
      and nb_cycles = ref 0
      in
      global :=
        M.merge (fun _x' oglobal olocal -> match oglobal, olocal with
            | None, Some _ ->
              changed := true;
              incr nb_new_states;
              incr nb_explored_states;
              olocal
            | Some _, None ->
              incr nb_states;
              oglobal
            | Some ({ visited = _; status = status_g }, g),
              Some ({ visited = _; status = status_l }, l) ->
              incr nb_states;
              incr nb_explored_states;
              let status = join_status status_g status_l in
              begin match status with
                | NoCycle -> ()
                | MayCycle -> incr nb_cycles
              end;
              if D.leq l g
              then
                (** local value is smaller: we have a stable point *)
                Some ({ visited = false; status }, g)
              else begin
                (** the point is not stable *)
                changed := true;
                incr nb_unstable_states;
                let join = match status with
                  | NoCycle -> D.join
                  | MayCycle -> D.widen
                in
                Some ({ visited = false; status }, join g l)
              end
            | None, None -> assert false
          )
          !global
          !local;
      begin
        if debug
        then
          Log.eprintf "Iteration %i: states=%i explored=%i unstable=%i new=%i cycles=%i@."
            !nb_iterations
            !nb_states
            !nb_explored_states
            !nb_unstable_states
            !nb_new_states
            !nb_cycles;
      end;
      if not !changed
      then
        (** the graph has not changed: we have a post-fixpoint *)
        y
      else
        (** we have not converged yet: iterate more *)
        stabilize_from x
    in
    (fun x ->
       let res = stabilize_from x in
       let nb_nodes = M.cardinal !global
       and nb_edges = Edges.cardinal !edges
       and max_fanout = Edges.max_fanout !edges in
       begin
         if debug
         then
           Log.eprintf
             (* "Elements in cache:@.@[%a@]@.\
              * " *)
             "Total number of explored states: %i@.\
              Total number of edges: %i@.\
              Max fanout: %i@.\
              Total number of iterations: %i@."
             (* pp_cache !global *)
             nb_nodes
             nb_edges
             max_fanout
             !nb_iterations
       end;
       let stats = {
         Statistics.states = nb_nodes;
         edges = nb_edges;
         max_fanout = max_fanout;
         iterations = !nb_iterations;
       }
       in
       begin
         if not transient
         then reset global
       end;
       nb_iterations := 0;
       (res, stats))
end
